from textwrap import dedent
from typing import Callable, Optional, Type, TypedDict
from langchain import BasePromptTemplate, PromptTemplate, LLMChain
from langchain.chat_models import ChatOpenAI
from genworlds.agents.tree_agent.brains.brain import Brain
from genworlds.agents.tree_agent.prompts.execution_generator_prompt import (
    ExecutionGeneratorPrompt,
)

from langchain.output_parsers.openai_functions import JsonKeyOutputFunctionsParser, JsonOutputFunctionsParser
from langchain.chains.openai_functions.utils import get_llm_kwargs

class ZeroShotBrain(Brain):
    """This brain generates one output and returns it."""

    def __init__(
        self,
        openai_api_key: str,
        prompt_template_class: Type[BasePromptTemplate],
        llm_params: list[str],
        generator_role_prompt: str,
        generator_results_prompt: str,
        output_parameter_generator: Callable[[dict], dict[str, dict]],
        model_name="gpt-4",
        temperature=0.7,
        verbose=False,
        request_timeout=120,
    ):
        self.verbose = verbose

        self.prompt_template_class = prompt_template_class
        self.llm_params = llm_params

        self.output_parameter_generator = output_parameter_generator

        self.llm = ChatOpenAI(
            temperature=temperature,
            openai_api_key=openai_api_key,
            model_name=model_name,
            request_timeout=request_timeout,
        )

        self.gen_prompt = prompt_template_class(
            token_counter=self.llm.get_num_tokens,
            input_variables=llm_params,
            ai_role=dedent(generator_role_prompt),
            response_instruction=dedent(generator_results_prompt),
        )

    def gen_thoughts(
        self,
        llm_params: dict,
    ):
        output_parameters = self.output_parameter_generator(llm_params)

        generator_function = {
            "name": "generate_output",
            "description": f"Generates the desired output parameters.",
            "parameters": {
                "type": "object",
                "properties": output_parameters,
                "required": list(output_parameters.keys()),
            },
        }


        llm_kwargs = get_llm_kwargs(generator_function)
        output_parser = JsonKeyOutputFunctionsParser(key_name="options")

        if self.verbose:
            print(llm_kwargs)

        gen_llm_chain = LLMChain(
            prompt=self.gen_prompt,
            llm=self.llm,
            llm_kwargs=llm_kwargs,
            output_parser=output_parser,
            verbose=self.verbose,
        )

        # prepare the input variables
        response = gen_llm_chain.run(
            **llm_params,
        )

        if self.verbose:
            print("Generated: " + str(response))

        return response

    def run(self, llm_params: dict):
        return self.gen_thoughts(llm_params)



from __future__ import annotations

import os
import json
from functools import partial

from prompt_toolkit.layout import HSplit, VSplit
from prompt_toolkit.layout.containers import Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.widgets import RadioList
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.widgets import Button
from prompt_toolkit.layout import WindowAlign
from prompt_toolkit.layout.dimension import D

import genworlds.interfaces as interfaces


def loading_socket_servers(genworlds_path: str):
    ## gather all the ws running locally if is empty display the empty message
    file_path = os.path.join(genworlds_path, "active_socket_servers_list.txt")
    values = []

    with open(file_path, "r") as f:
        config = f.readlines()
    for line in list(set(config)):
        values.append((line, line))

    socket_servers_menu = RadioList(
        values=values,
    )
    return socket_servers_menu


def loading_configuration_files(genworlds_path: str):
    config_path = os.path.join(genworlds_path, "cli_configurations")
    values = []

    for file in os.listdir(config_path):
        file_path = os.path.join(config_path, file)
        with open(file_path, "r") as f:
            config = json.load(f)
        values.append((file_path, config["name"]))

    screen_config_menu = RadioList(
        values=values,
    )
    return screen_config_menu


# launch CLI button init the client threaded pointing to the selected ws server and switch the CLI layout
def launch_cli_button_handler(
    cli: interfaces.CLI, socket_servers_menu: RadioList, screen_config_menu: RadioList
):
    cli.server_url = socket_servers_menu.current_value
    with open(screen_config_menu.current_value, "r") as f:
        cli.configuration = json.load(f)
    cli.initialize_dynamic_layout()


def initial_setup_layout_screen(cli: interfaces.CLI, genworlds_path: str):
    title = Window(
        height=D.exact(3),
        content=FormattedTextControl(
            text=HTML(
                '<style fg="#ffb86c">\n🧬🌍 GenWorlds CLI Configuration Screen\n</style>'
            )
        ),
        align=WindowAlign.CENTER,
    )
    socket_servers_menu = loading_socket_servers(genworlds_path=genworlds_path)
    screen_config_menu = loading_configuration_files(genworlds_path=genworlds_path)

    socket_server_layout = HSplit(
        [
            Window(
                content=FormattedTextControl(
                    text="Select the URL of the World WS Server to connect to:",
                    focusable=True,
                ),
                height=1,
                style="reverse",
            ),
            socket_servers_menu,
        ]
    )

    screen_config_layout = HSplit(
        [
            Window(
                content=FormattedTextControl(
                    text="Select CLI Config File or start a new one:", focusable=True
                ),
                height=1,
                style="reverse",
            ),
            screen_config_menu,
        ],
    )

    main_container = VSplit([socket_server_layout, screen_config_layout])

    launch_cli_button = Button(
        "Launch CLI",
        handler=partial(
            launch_cli_button_handler, cli, socket_servers_menu, screen_config_menu
        ),
    )

    # Update the CLI layout
    cli.application.layout.container = HSplit(
        [
            title,
            main_container,
            Window(height=5),
            launch_cli_button,
            Window(height=1),
        ]
    )

    # Update initial focus
    cli.application.layout.focus(socket_servers_menu)

    # Update the CLI keybindings
    @cli.kb.add("right")
    def _(event):
        if cli.application.layout.has_focus(screen_config_menu):
            cli.application.layout.focus(launch_cli_button)
        else:
            cli.application.layout.focus(screen_config_menu)

    @cli.kb.add("left")
    def _(event):
        # focus the left RadioList
        if cli.application.layout.has_focus(launch_cli_button):
            cli.application.layout.focus(screen_config_menu)
        else:
            cli.application.layout.focus(socket_servers_menu)



import threading
from typing import Callable, List, Optional
from pydantic import BaseModel, Field
from datetime import datetime

from genworlds.sockets.world_socket_client import WorldSocketClient


class Event(BaseModel):
    event_type: str
    description: str
    summary: Optional[str]
    created_at: datetime
    sender_id: str
    target_id: Optional[str] = Field(
        description="ID of the entity that handles the event"
    )


class WebsocketEventHandler:
    listeners: dict[str, set]
    event_classes: dict[str, type[Event]]

    def __init__(
        self,
        id,
        event_class_listener_pairs: List[tuple[type[Event], Callable]] = None,
        websocket_url: str = "ws://127.0.0.1:7456/ws",
    ):
        if event_class_listener_pairs is None:
            event_class_listener_pairs = []

        self.listeners = {}
        self.event_classes = {}

        self.id = id
        self.register_event_listeners(event_class_listener_pairs)
        self.world_socket_client = WorldSocketClient(
            process_event=self.process_event, url=websocket_url
        )

    def register_event_listeners(
        self, event_class_listener_pairs: List[tuple[type[Event], Callable]]
    ):
        for event_class, event_listener in event_class_listener_pairs:
            self.register_event_listener(event_class, event_listener)

    def register_event_listener(
        self, event_class: type[Event], event_listener: Callable
    ):
        event_type = event_class.__fields__["event_type"].default
        self.event_classes[event_type] = event_class

        if event_type not in self.listeners:
            self.listeners[event_type] = set()

        self.listeners[event_type].add(event_listener)

    def unregister_event_listener(
        self, event_class: type[Event], event_listener: Callable
    ):
        if event_class.event_type in self.listeners:
            self.listeners[event_class.event_type].remove(event_listener)

            if len(self.listeners[event_class.event_type]) == 0:
                del self.listeners[event_class.event_type]

    def process_event(self, event):
        if event["event_type"] in self.listeners and (
            event["target_id"] == None or event["target_id"] == self.id
        ):
            parsed_event = self.event_classes[event["event_type"]].parse_obj(event)

            for listener in self.listeners[event["event_type"]]:
                listener(parsed_event)

    def send_event(self, event_class: type[Event], **kwargs):
        event = event_class(sender_id=self.id, created_at=datetime.now(), **kwargs)

        self.world_socket_client.send_message(event.json())

    def launch_websocket_thread(self):
        threading.Thread(
            target=self.world_socket_client.websocket.run_forever,
            name=f"{self.id} Thread",
            daemon=True,
        ).start()



from __future__ import annotations

import genworlds.interfaces as interfaces
from genworlds.interfaces.cli.formatted_buffer import html_from_json_format


def process_event_router(self: interfaces.CLI, ws_json_message):
    window_width = self.application.layout.current_window.render_info.window_width - 5

    if len(self.screens) > 0:
        for screen_name in self.screens:
            screen = self.screens[screen_name]
            if screen["screen_type"] == "event_history":
                if ws_json_message["event_type"] in [
                    ev["event_type"] for ev in screen["tracked_events"]
                ]:  # modify buffers accordingly
                    screen["buffer"].text += event_history_formatter(
                        screen, ws_json_message, max_width=window_width
                    )
            elif screen["screen_type"] == "entities_current_state":
                screen["buffer"].text += entities_current_state_formatter(
                    screen, ws_json_message, max_width=window_width
                )


def event_history_formatter(screen, message, max_width=100):
    text = ""
    event_type = message["event_type"]
    for et in screen["tracked_events"]:
        if et["event_type"] != event_type:
            continue

        field_filters = et["filters"]
        for field in et["fields_to_display"]:
            field_name = field["name"]
            field_format = field["format"]

            if field_name not in message:
                continue

            if len(field_filters) > 0:
                for filter in field_filters:
                    field = filter["field"]
                    value = filter["value"]
                    if message[field] == value:
                        text += html_from_json_format(
                            field_format,
                            f"{field_name}: {message[field_name]}",
                            max_width=max_width,
                        )
                        text += "\n"
                        continue
            else:
                text += html_from_json_format(
                    field_format,
                    f"{field_name}: {message[field_name]}",
                    max_width=max_width,
                )
                text += "\n"
                continue

    return text + "\n"


def entities_current_state_formatter(screen, message, max_width=100):
    return ""



from uuid import uuid4
from typing import Generic, Type, TypeVar

import uvicorn

from genworlds.agents.yeager_autogpt.agent import YeagerAutoGPT
from genworlds.objects.base_object import BaseObject
from genworlds.events.websocket_event_handler import WebsocketEventHandler
from genworlds.events.basic_events import (
    AgentGetsNearbyEntitiesEvent,
    AgentGivesObjectToAgentEvent,
    AgentSpeaksWithAgentEvent,
    EntityRequestWorldStateUpdateEvent,
    EntityWorldStateUpdateEvent,
    WorldSendsNearbyEntitiesEvent,
    WorldSendsAllEntitiesEvent,
    WorldSendsSchemasEvent,
)
from genworlds.worlds.base_world_entity import EntityTypeEnum, BaseWorldEntity

WorldEntityType = TypeVar("WorldEntityType", bound=BaseWorldEntity)


class BaseWorld(Generic[WorldEntityType], WebsocketEventHandler):
    subconscious_event_classes: set[str]
    entities: dict[str, WorldEntityType]
    entity_schemas: dict[str, dict]
    world_entity_constructor: Type[WorldEntityType]

    def __init__(
        self,
        world_entity_constructor: Type[WorldEntityType],
        name: str,
        description: str,
        id: str = None,
        websocket_url: str = "ws://127.0.0.1:7456/ws",
    ):
        self.id = id if id else str(uuid4())
        self.name = name
        self.description = description
        self.world_entity_constructor = world_entity_constructor
        self.subconscious_event_classes = {
            AgentGetsNearbyEntitiesEvent,
            EntityRequestWorldStateUpdateEvent,
        }

        super().__init__(self.id, websocket_url=websocket_url)

        self.entities = {}
        self.entity_schemas = {}

        self.register_event_listeners(
            [
                (
                    AgentGetsNearbyEntitiesEvent,
                    self.agent_gets_nearby_entities_listener,
                ),
                (
                    EntityRequestWorldStateUpdateEvent,
                    self.entity_request_world_state_update_event_listener,
                ),
                (
                    AgentGivesObjectToAgentEvent,
                    self.agent_gives_object_to_agent_listener,
                ),
            ]
        )

    def get_agent_by_id(self, agent_id: str) -> WorldEntityType:
        entity = self.entities.get(agent_id)
        if entity is None:
            raise Exception(f"Agent with id {agent_id} does not exist.")
        if entity.entity_type != EntityTypeEnum.AGENT:
            raise Exception(f"Entity with id {agent_id} is not an agent.")
        return entity

    def get_object_by_id(self, object_id: str) -> WorldEntityType:
        entity = self.entities.get(object_id)
        if entity is None:
            raise Exception(f"Object with id {object_id} does not exist.")
        if entity.entity_type != EntityTypeEnum.OBJECT:
            raise Exception(f"Entity with id {object_id} is not an object.")
        return entity

    ## Event Handlers

    def get_nearby_entities(self, entity_id: str) -> WorldEntityType:
        # Check if reference entity exists
        self.entities.get(entity_id)

        # All entities
        return self.entities

    def agent_gets_nearby_entities_listener(self, event: AgentGetsNearbyEntitiesEvent):
        self.emit_world_sends_nearby_entities(agent_id=event.agent_id)

    def emit_world_sends_nearby_entities(self, agent_id: str):
        nearby_entities = self.get_nearby_entities(agent_id)

        self.send_event(
            WorldSendsNearbyEntitiesEvent,
            target_id=agent_id,
            nearby_entities=nearby_entities,
        )

    def get_agent_world_state_prompt(self, agent_id: str) -> str:
        agent_entity = self.get_agent_by_id(agent_id)

        world_state_prompt = (
            f"You are an agent in the world.\n" f"Your id is {agent_entity.id}.\n"
        )

        return world_state_prompt

    def emit_agent_world_state(self, agent_id):
        world_state_prompt = self.get_agent_world_state_prompt(agent_id=agent_id)

        self.send_event(
            EntityWorldStateUpdateEvent,
            target_id=agent_id,
            entity_world_state=world_state_prompt,
        )

    def entity_request_world_state_update_event_listener(
        self, event: EntityRequestWorldStateUpdateEvent
    ):
        self.world_sends_schemas()
        self.world_sends_all_entities()
        self.emit_agent_world_state(agent_id=event.sender_id)
        self.emit_world_sends_nearby_entities(agent_id=event.sender_id)

    def agent_gives_object_to_agent_listener(self, event: AgentGivesObjectToAgentEvent):
        if event.object_id not in self.entities:
            raise Exception(f"Object with id {event.object_id} does not exist.")

        if event.recipient_agent_id not in self.entities:
            raise Exception(f"Agent with id {event.recipient_agent_id} does not exist.")

        if self.entities[event.object_id].held_by != event.sender_id:
            raise Exception(
                f"Object with id {event.object_id} is not held by agent with id {event.sender_id}."
            )

        self.entities[event.object_id].held_by = event.recipient_agent_id

        self.emit_world_sends_nearby_entities(agent_id=event.sender_id)
        self.emit_world_sends_nearby_entities(agent_id=event.recipient_agent_id)

    def world_sends_schemas(self):
        schemas = self.entity_schemas.copy()
        # Add world schemas
        events = {}
        for event_type, event in self.event_classes.items():
            if event not in self.subconscious_event_classes:
                events[event_type] = event.schema()

        schemas["World"] = events
        self.send_event(
            WorldSendsSchemasEvent,
            world_name=self.name,
            world_description=self.description,
            schemas=schemas,
        )

    def world_sends_all_entities(self):
        all_entities = self.entities.copy()
        self.send_event(WorldSendsAllEntitiesEvent, all_entities=[all_entities])

    def register_agent(self, agent: YeagerAutoGPT, **kwargs: WorldEntityType):
        class_name = agent.__class__.__name__

        if class_name not in self.entity_schemas:
            events = {}

            # for event_class in agent.interesting_events:
            #     events[event_class.__fields__['event_type'].default] = event_class.schema()

            self.entity_schemas[class_name] = events

        entity: WorldEntityType = self.world_entity_constructor(
            id=agent.id,
            entity_type=EntityTypeEnum.AGENT,
            entity_class=class_name,
            name=agent.name,
            description=agent.description,
            events=self.entity_schemas[class_name],
            **kwargs,
        )

        print("Registered agent", entity)

        self.entities[entity.id] = entity

    def register_object(self, obj: BaseObject, **kwargs: WorldEntityType):
        class_name = obj.__class__.__name__

        if class_name not in self.entity_schemas:
            events = {}
            for event_type, event in obj.event_classes.items():
                events[event_type] = event.schema()

            self.entity_schemas[class_name] = events

        entity: WorldEntityType = self.world_entity_constructor(
            id=obj.id,
            entity_type=EntityTypeEnum.OBJECT,
            entity_class=class_name,
            name=obj.name,
            description=obj.description,
            **kwargs,
        )

        print("Registered object", entity)

        self.entities[entity.id] = entity



from langchain import PromptTemplate, LLMChain
from langchain.chat_models import ChatOpenAI


class MemorySummarizer:
    def __init__(self, openai_api_key: str, model_name: str = "gpt-3.5-turbo"):
        self.summary_template = """
The following is an event in JSON format:
{memory}
Summarize what happened in one line.
Follow the general format of - who did what, to whom if applicable.
        """

        self.summary_prompt = PromptTemplate(
            template=self.summary_template,
            input_variables=[
                "memory",
            ],
        )

        self.chat = ChatOpenAI(
            temperature=0, model_name=model_name, openai_api_key=openai_api_key
        )
        self.chain = LLMChain(llm=self.chat, prompt=self.summary_prompt)

    def summarize(
        self,
        memory: str,
    ) -> str:
        """
        Summarize the memory in one line.
        """
        return self.chain.run(memory=memory)



from enum import Enum

from pydantic import BaseModel


class EntityTypeEnum(str, Enum):
    AGENT = "AGENT"
    OBJECT = "OBJECT"
    WORLD = "WORLD"


class BaseWorldEntity(BaseModel):
    id: str
    entity_type: EntityTypeEnum
    entity_class: str
    name: str
    description: str
    held_by: str = None



import time
import os
import threading

from prompt_toolkit.application import Application
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.layout.containers import Window
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.styles import Style

from genworlds.sockets.world_socket_client import WorldSocketClient
from genworlds.interfaces.cli.event_processors import process_event_router
from genworlds.interfaces.cli.initial_setup_layout_screen import (
    initial_setup_layout_screen,
)
from genworlds.interfaces.cli.render_main_screen import render_main_screen

if os.getenv("IS_REPLIT") == "True":
    GENWORLDS_CONFIG_PATH = os.path.join(os.getcwd(), ".replit_config")
else:
    GENWORLDS_CONFIG_PATH = os.path.join(os.path.expanduser("~"), ".genworlds")


class CLI:
    def __init__(self):
        self.ws_client = WorldSocketClient(self.process_event, log_level="ERROR")
        self.kb = KeyBindings()
        self.configuration = None
        self.server_url = None
        self.screens = {}
        self.current_screen = None
        self.selected_screen = None

        @self.kb.add("c-c", eager=True)
        def exit_app(event):
            event.app.exit()

        self.layout = Layout(Window())
        self.application = Application(
            layout=self.layout,
            key_bindings=self.kb,
            mouse_support=True,
            full_screen=True,
        )

        initial_setup_layout_screen(self, GENWORLDS_CONFIG_PATH)

    def process_event(self, ws_json_message):
        process_event_router(self, ws_json_message)

    def initialize_dynamic_layout(self):
        for screen in self.configuration["screens"]:
            self.screens[screen["name"]] = {}
            self.screens[screen["name"]]["has_input"] = screen["has_input"]
            self.screens[screen["name"]]["screen_type"] = screen["screen_type"]
            self.screens[screen["name"]]["buffer"] = Buffer()
            if self.screens[screen["name"]]["screen_type"] == "event_history":
                self.screens[screen["name"]]["tracked_events"] = screen[
                    "tracked_events"
                ]
            elif (
                self.screens[screen["name"]]["screen_type"] == "entities_current_state"
            ):
                self.screens[screen["name"]]["tracked_entities"] = screen[
                    "tracked_entities"
                ]

        render_main_screen(self)

    def run(self):
        time.sleep(2)
        threading.Thread(
            target=self.ws_client.websocket.run_forever,
            name=f"CLI Socket Client Thread",
            daemon=True,
        ).start()
        self.application.run()



from genworlds.events.websocket_event_handler import Event


class AgentGetsNearbyEntitiesEvent(Event):
    event_type = "agent_gets_nearby_entities_event"
    description = "Get all entities near an agent."


class WorldSendsNearbyEntitiesEvent(Event):
    event_type = "world_sends_nearby_entities_event"
    description = "Send all nearby entities to an agent."
    nearby_entities: list


class WorldSendsAllEntitiesEvent(Event):
    event_type = "world_sends_all_entities_event"
    description = "Send all entities."
    all_entities: list


class AgentGetsObjectInfoEvent(Event):
    event_type = "agent_gets_object_info_event"
    description = "Get info about an object."


class AgentGetsAgentInfoEvent(Event):
    event_type = "agent_gets_agent_info_event"
    description = "Get info about an agent."


class ObjectSendsInfoToAgentEvent(Event):
    event_type = "object_sends_info_to_agent_event"
    description = "Send info about an object to an agent."
    object_id: str
    object_name: str
    object_description: str
    possible_events: list[Event]


class AgentSpeaksWithAgentEvent(Event):
    event_type = "agent_speaks_with_agent_event"
    description = "An agent speaks with another agent."
    message: str


class WorldSendsSchemasEvent(Event):
    event_type = "world_sends_schemas_event"
    description = "The world sends the possible interactions to all the agents."
    world_name: str
    world_description: str
    schemas: dict[str, dict]


class EntityRequestWorldStateUpdateEvent(Event):
    event_type = "entity_request_world_state_update_event"
    description = "Request the latest world state update for an entity."


class EntityWorldStateUpdateEvent(Event):
    event_type = "entity_world_state_update_event"
    description = "Latest world state update for an entity."
    entity_world_state: str


class AgentGivesObjectToAgentEvent(Event):
    event_type = "agent_gives_object_to_agent_event"
    description = "Give an object from your inventory to another agent. Only the holder of an item can use this event, you cannot use this event to request an item. Target id must be the id of the world."
    object_id: str
    recipient_agent_id: str



import json
import time
from typing import Any, Callable, List, Optional

from pydantic import BaseModel

from genworlds.agents.yeager_autogpt.prompt_generator import get_prompt
from langchain.prompts.chat import (
    BaseChatPromptTemplate,
)
from langchain.schema import BaseMessage, HumanMessage, SystemMessage
from langchain.tools.base import BaseTool
from langchain.vectorstores.base import VectorStoreRetriever
from langchain.vectorstores import VectorStore

from genworlds.agents.memory_processors.nmk_world_memory import NMKWorldMemory


class NavigationGeneratorPrompt(BaseChatPromptTemplate, BaseModel):
    ai_role: str
    response_instruction: str
    token_counter: Callable[[str], int]
    send_token_limit: int = 8192

    basic_template = """
# Basic rules
{ai_role}

## Your Goals
{goals}

## World State
{agent_world_state}
        """

    def construct_full_prompt(self, agent_world_state: str, goals: List[str]) -> str:
        return self.basic_template.format(
            ai_role=self.ai_role,
            goals="\n".join([f"{i+1}. {goal}" for i, goal in enumerate(goals)]),
            agent_world_state=agent_world_state,
        )

    def format_messages(self, **kwargs: Any) -> List[BaseMessage]:
        kwargs = {key: kwargs[key] for key in self.input_variables}

        messages: List[BaseMessage] = []

        base_prompt = SystemMessage(
            content=self.construct_full_prompt(
                kwargs["agent_world_state"], kwargs["goals"]
            )
        )
        messages.append(base_prompt)
        used_tokens = self.token_counter(base_prompt.content)

        if "previous_thoughts" in kwargs and len(kwargs["previous_thoughts"]) > 0:
            previous_thoughts_prompt = f"## Previous thoughts:\n"
            for entity in kwargs["previous_thoughts"]:
                previous_thoughts_prompt += f"- {json.dumps(entity)}\n"

            previous_thoughts_message = SystemMessage(content=previous_thoughts_prompt)
            messages.append(previous_thoughts_message)
            used_tokens += self.token_counter(previous_thoughts_message.content)

        time_prompt = SystemMessage(
            content=f"The current time and date is {time.strftime('%c')}"
        )
        messages.append(time_prompt)
        used_tokens = self.token_counter(base_prompt.content) + self.token_counter(
            time_prompt.content
        )

        inventory = kwargs["inventory"]
        inventory_prompt = f"## Inventory:\n"
        if len(inventory) > 0:
            for entity in inventory:
                inventory_prompt += f"- {json.dumps(entity)}\n"
        else:
            inventory_prompt += f"You have no items in your inventory.\n"
        inventory_message = SystemMessage(content=inventory_prompt)
        messages.append(inventory_message)
        used_tokens += self.token_counter(inventory_message.content)

        nearby_entities = kwargs["nearby_entities"]
        nearby_entities_prompt = f"## Nearby entities: \n"
        if len(nearby_entities) > 0:
            for entity in nearby_entities:
                nearby_entities_prompt += f"{json.dumps(entity)}\n"
        else:
            nearby_entities_prompt += f"There are no entities near you.\n"
        nearby_entities_message = SystemMessage(content=nearby_entities_prompt)
        messages.append(nearby_entities_message)
        used_tokens += self.token_counter(nearby_entities_message.content)

        if "relevant_commands" in kwargs and len(kwargs["relevant_commands"]) > 0:
            relevant_commands =  map(lambda c: c["string_short"], kwargs["relevant_commands"].values())
            relevant_commands_prompt = (
                f"You can perform the following actions with the entities nearby:\n"
            )
            for command in relevant_commands:
                relevant_commands_prompt += f"{command}\n"
            relevant_commands_message = SystemMessage(content=relevant_commands_prompt)
            messages.append(relevant_commands_message)
            used_tokens += self.token_counter(relevant_commands_message.content)

        if "plan" in kwargs and kwargs["plan"] is not None:
            plan = "\n".join([f"{i+1}. {goal}" for i, goal in enumerate(kwargs["plan"])]) 
            plan_message = SystemMessage(content=f"## Your Previous Plan:\n{plan}")
            messages.append(plan_message)
            used_tokens += len(plan_message.content)

            similarity_query_key = kwargs["plan"]
        else:
            similarity_query_key = kwargs["goals"]

        personality_db: VectorStore = kwargs["personality_db"]
        if personality_db is not None:
            past_statements = list(
                map(
                    lambda d: d.page_content,
                    personality_db.similarity_search(str(similarity_query_key)),
                )
            )
            if len(past_statements) > 0:
                past_statements_bullet_list = "\n".join(map(lambda s: f'- {s}', past_statements))
                past_statements_format = f"You have said the following things on this topic in the past:\n{past_statements_bullet_list}\n\n"
                personality_message = SystemMessage(content=past_statements_format)
                messages.append(personality_message)
                used_tokens += len(personality_message.content)

        if "memory" in kwargs and kwargs["memory"] is not None:
            memory: NMKWorldMemory = kwargs["memory"]
            if len(memory.world_events) > 0:
                if len(memory.world_events) % 5 == 0:
                    memory.create_full_summary()
                relevant_memory = memory.get_event_stream_memories(
                    query=similarity_query_key, summarized=False
                )  # TODO: add token limitations
                relevant_memory_tokens = self.token_counter(relevant_memory)
                memory_message = SystemMessage(content=relevant_memory)
                messages.append(memory_message)
                used_tokens += relevant_memory_tokens

        instruction = HumanMessage(content=self.response_instruction.format(**kwargs))
        messages.append(instruction)

        return messages



import json
from textwrap import dedent
from typing import Callable, Optional, Type, TypedDict
from langchain import BasePromptTemplate, PromptTemplate, LLMChain
from langchain.chat_models import ChatOpenAI
from genworlds.agents.tree_agent.brains.brain import Brain
from genworlds.agents.tree_agent.prompts.execution_generator_prompt import (
    ExecutionGeneratorPrompt,
)

from langchain.output_parsers.openai_functions import JsonKeyOutputFunctionsParser, JsonOutputFunctionsParser
from langchain.chains.openai_functions.utils import get_llm_kwargs


class SingleEvalBrain(Brain):
    """This brain generates a number of thoughts and passes them all to the evaluator, which selects one of them."""

    def __init__(
        self,
        openai_api_key: str,
        prompt_template_class: Type[BasePromptTemplate],
        llm_params: list[str],
        generator_role_prompt: str,
        generator_results_prompt: str,
        evaluator_role_prompt: str,
        evaluator_results_prompt: str,
        n_of_thoughts: int,
        output_parameter_generator: Callable[[dict], dict[str, dict]],
        model_name="gpt-4",
        temperature=0.7,
        verbose=False,
        request_timeout=120,
    ):
        self.n_of_thoughts = n_of_thoughts
        self.verbose = verbose

        self.prompt_template_class = prompt_template_class
        
        self.output_parameter_generator = output_parameter_generator

        self.llm = ChatOpenAI(
            temperature=temperature,
            openai_api_key=openai_api_key,
            model_name=model_name,
            request_timeout=request_timeout,
        )

        self.gen_prompt = prompt_template_class(
            token_counter=self.llm.get_num_tokens,
            input_variables=[
                "previous_thoughts",
                "num_thoughts",
            ]
            + llm_params,
            ai_role=dedent(generator_role_prompt),
            response_instruction=dedent(generator_results_prompt),
        )
        
        self.eval_prompt = prompt_template_class(
            token_counter=self.llm.get_num_tokens,
            input_variables=[
                "thought_to_evaluate",
            ]
            + llm_params,
            ai_role=dedent(evaluator_role_prompt),
            response_instruction=dedent(evaluator_results_prompt),
        )
        

    def gen_thoughts(
        self,
        previous_thoughts,
        num_thoughts: int,
        llm_params: dict,
    ):
        generator_function = {
            "name": "generate_options",
            "description": f"Generates {num_thoughts} options for the agent to choose from.",
            "parameters": {
                "type": "object",
                "properties": {
                    "options": {
                        "type": "array",
                        "minItems": num_thoughts,
                        "maxItems": num_thoughts,
                        "items": {
                            "type": "object",
                            "properties": self.output_parameter_generator(llm_params),
                        }
                    },
                },
                "required": ["options"],
            },
        }


        llm_kwargs = get_llm_kwargs(generator_function)
        output_parser = JsonKeyOutputFunctionsParser(key_name="options")

        if self.verbose:
            print(llm_kwargs)

        gen_llm_chain = LLMChain(
            prompt=self.gen_prompt,
            llm=self.llm,
            llm_kwargs=llm_kwargs,
            output_parser=output_parser,
            verbose=self.verbose,
        )

        # prepare the input variables
        response = gen_llm_chain.run(
            num_thoughts=num_thoughts,
            previous_thoughts=previous_thoughts,
            **llm_params,
        )

        if self.verbose:
            print("Generated: " + str(response))

        return response

    def eval_thoughts(
        self,
        thoughts_to_evaluate: list[dict],
        llm_params: dict,
    ):
        output_parameters = self.output_parameter_generator(llm_params)

        evaluator_function = {
            "name": "generate_output",
            "description": "Generate the final output of the agent.",
            "parameters": {
                "type": "object",
                "properties": output_parameters,                        
                "required": list(output_parameters.keys()),
            },
        }

        llm_kwargs = get_llm_kwargs(evaluator_function)
        output_parser = JsonOutputFunctionsParser()

        eval_llm_chain = LLMChain(
            prompt=self.eval_prompt,
            llm=self.llm,
            llm_kwargs=llm_kwargs,
            output_parser=output_parser,
            verbose=self.verbose,
        )

        response = eval_llm_chain.run(
            thought_to_evaluate=json.dumps(thoughts_to_evaluate),
            **llm_params,
        )

        if self.verbose:
            print("Evaluated: " + str(response))

        return response


    def run(self, llm_params: dict):
        thoughts = self.gen_thoughts("", self.n_of_thoughts, llm_params)

        if len(thoughts) == 1:
            return thoughts[0]

        best_thought = self.eval_thoughts(thoughts, llm_params)

        return best_thought



import json
import time
from typing import Any, Callable, List, Optional

from pydantic import BaseModel

from langchain.schema import BaseMessage, HumanMessage, SystemMessage
from langchain.tools.base import BaseTool
from langchain.vectorstores.base import VectorStoreRetriever
from langchain.vectorstores import VectorStore
from langchain.prompts.chat import (
    BaseChatPromptTemplate,
)

from genworlds.agents.memory_processors.nmk_world_memory import NMKWorldMemory


class ExecutionGeneratorPrompt(BaseChatPromptTemplate, BaseModel):
    ai_role: str
    response_instruction: str
    token_counter: Callable[[str], int]
    send_token_limit: int = 8192

    basic_template = """
# Basic rules
{ai_role}

## Your Goals
{goals}

## World State
{agent_world_state}
        """

    def construct_full_prompt(self, agent_world_state: str, goals: List[str]) -> str:
        return self.basic_template.format(
            ai_role=self.ai_role,
            goals="\n".join([f"{i+1}. {goal}" for i, goal in enumerate(goals)]),
            agent_world_state=agent_world_state,
        )

    def format_messages(self, **kwargs: Any) -> List[BaseMessage]:
        kwargs = {key: kwargs[key] for key in self.input_variables}

        messages: List[BaseMessage] = []

        base_prompt = SystemMessage(
            content=self.construct_full_prompt(
                kwargs["agent_world_state"], kwargs["goals"]
            )
        )
        messages.append(base_prompt)
        used_tokens = self.token_counter(base_prompt.content)

        if "previous_thoughts" in kwargs and len(kwargs["previous_thoughts"]) > 0:
            previous_thoughts_prompt = f"## Previous thoughts:\n"
            for entity in kwargs["previous_thoughts"]:
                previous_thoughts_prompt += f"- {json.dumps(entity)}\n"
            previous_thoughts_message = SystemMessage(content=previous_thoughts_prompt)
            messages.append(previous_thoughts_message)
            used_tokens += self.token_counter(previous_thoughts_message.content)

        if (
            "previous_brain_outputs" in kwargs
            and len(kwargs["previous_brain_outputs"]) > 0
        ):
            previous_brain_outputs_prompt = f"## Previous thoughts:\n"

            for entity in kwargs["previous_brain_outputs"]:
                previous_brain_outputs_prompt += f"- {json.dumps(entity)}\n"

            previous_brain_outputs_message = SystemMessage(
                content=previous_brain_outputs_prompt
            )
            messages.append(previous_brain_outputs_message)
            used_tokens += self.token_counter(previous_brain_outputs_message.content)

        time_prompt = SystemMessage(
            content=f"The current time and date is {time.strftime('%c')}"
        )
        messages.append(time_prompt)
        used_tokens = self.token_counter(base_prompt.content) + self.token_counter(
            time_prompt.content
        )

        if "inventory" in kwargs:
            inventory = kwargs["inventory"]
            inventory_prompt = f"## Inventory:\n"
            if len(inventory) > 0:
                for entity in inventory:
                    inventory_prompt += f"- {json.dumps(entity)}\n"
            else:
                inventory_prompt += f"You have no items in your inventory.\n"
            inventory_message = SystemMessage(content=inventory_prompt)
            messages.append(inventory_message)
            used_tokens += self.token_counter(inventory_message.content)

        if "nearby_entities" in kwargs:
            nearby_entities = kwargs["nearby_entities"]
            nearby_entities_prompt = f"## Nearby entities: \n"
            if len(nearby_entities) > 0:
                for entity in nearby_entities:
                    nearby_entities_prompt += f"{json.dumps(entity)}\n"
            else:
                nearby_entities_prompt += f"There are no entities near you.\n"
            nearby_entities_message = SystemMessage(content=nearby_entities_prompt)
            messages.append(nearby_entities_message)
            used_tokens += self.token_counter(nearby_entities_message.content)

        if "command_to_execute" in kwargs:
            command_to_execute = kwargs["command_to_execute"]["string_full"]
            command_to_execute_message = SystemMessage(
                content=f"## You have selected the following command to execute:\n{command_to_execute}"
            )
            messages.append(command_to_execute_message)
            used_tokens += len(command_to_execute_message.content)

        if "plan" in kwargs:
            plan = "\n".join([f"{i+1}. {goal}" for i, goal in enumerate(kwargs["plan"])]) 
            plan_message = SystemMessage(content=f"## Your Plan:\n{plan}")
            messages.append(plan_message)
            used_tokens += len(plan_message.content)   

        if "personality_db" in kwargs and kwargs["personality_db"] is not None:
            personality_db: VectorStore = kwargs["personality_db"]
            past_statements = [document.page_content for document in personality_db.similarity_search(str(kwargs["previous_brain_outputs"][-1]))]
            
            if len(past_statements) > 0:
                past_statements_bullet_list = "\n".join(map(lambda s: f'- {s}', past_statements))
                past_statements_format = f"You have said the following things on this topic in the past:\n{past_statements_bullet_list}\n\n"
                personality_message = SystemMessage(content=past_statements_format)
                messages.append(personality_message)
                used_tokens += len(personality_message.content)

        if "memory" in kwargs and kwargs["memory"] is not None:
            memory: NMKWorldMemory = kwargs["memory"]
            if len(memory.world_events) > 0:
                if len(memory.world_events) % 5 == 0:
                    memory.create_full_summary()
                relevant_memory = memory.get_event_stream_memories(
                    query=kwargs["previous_brain_outputs"][-1], summarized=False
                )  # TODO: add token limitations
                relevant_memory_tokens = self.token_counter(relevant_memory)
                memory_message = SystemMessage(content=relevant_memory)
                messages.append(memory_message)
                used_tokens += relevant_memory_tokens

        instruction = HumanMessage(content=self.response_instruction.format(**kwargs))
        messages.append(instruction)

        return messages



from __future__ import annotations

from prompt_toolkit.layout import HSplit
from prompt_toolkit.layout.containers import Window
from prompt_toolkit.layout.controls import FormattedTextControl, BufferControl
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.widgets import MenuContainer, MenuItem
from prompt_toolkit.layout import WindowAlign
from prompt_toolkit.layout.dimension import D
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout.margins import ScrollbarMargin

import genworlds.interfaces as interfaces
from genworlds.interfaces.cli.render_help_screen import render_help_screen
from genworlds.interfaces.cli.formatted_buffer import (
    FormatText,
)


def render_main_screen(cli: interfaces.CLI):
    title = Window(
        height=D.exact(3),
        content=FormattedTextControl(
            text=HTML('<style fg="#ffb86c">\n🧬🌍 GenWorlds CLI Main Screen\n</style>')
        ),
        align=WindowAlign.CENTER,
    )

    screens_menu = MenuContainer(
        body=Window(height=0),
        menu_items=[
            MenuItem(f"  {k}  ", handler=lambda k=k: switch_buffer(k))
            for k in cli.screens.keys()
        ],
    )
    screen_buffer = Window(
        BufferControl(
            buffer=cli.screens[list(cli.screens.keys())[0]]["buffer"],
            focusable=True,
            input_processors=[FormatText()],
        ),
        right_margins=[ScrollbarMargin(display_arrows=True)],
    )
    prompt_buffer = Window(height=1)  # TODO: Create a reasonable prompt_buffer
    main_container = HSplit(
        [
            screens_menu,
            Window(height=1),
            screen_buffer,
            Window(height=1, char="-", style="class:line"),
            prompt_buffer,
        ]
    )
    # Define the commands menu
    commands_menu = MenuContainer(
        body=Window(height=0),
        menu_items=[
            MenuItem(
                "^W - Write Buffers To Disk   "
            ),  # , handler=lambda: write_buffers(), shortcut=Keys.ControlW),
            MenuItem(
                "^H - Help   "
            ),  # , handler=lambda: render_help_screen(cli), shortcut=Keys.ControlH),
            MenuItem(
                "^C - Exit   "
            ),  # , handler=lambda: cli.application.exit(), shortcut=Keys.ControlC),
        ],
    )

    # Update the CLI layout
    cli.application.layout.container = HSplit(
        [
            title,
            main_container,
            Window(height=1, char="-", style="class:line"),
            commands_menu,
            Window(height=1),
        ],
    )

    # Update initial focus
    cli.application.layout.focus(screens_menu)

    # Update the CLI keybindings
    @cli.kb.add(Keys.ControlW, eager=True)
    def _(event):
        write_buffers()

    @cli.kb.add(Keys.ControlH, eager=True)
    def _(event):
        render_help_screen(cli)

    def switch_buffer(screen_name):
        screen_buffer.content = BufferControl(
            buffer=cli.screens[screen_name]["buffer"],
            input_processors=[FormatText()],
            focusable=True,
        )
        if cli.screens[screen_name]["has_input"]:
            """TODO: If the screen has input, focus the prompt buffer and create a reasonable prompt_buffer"""
        cli.application.layout.focus(screens_menu)

    def write_buffers():
        # for screen_name in cli.screens.keys():
        #     with open(screen_name, "w") as f:
        #         f.write(cli.screens[screen_name]["buffer"].text)
        cli.application.layout.focus(screens_menu)



import json
import time
from typing import Any, Callable, List, Optional

from pydantic import BaseModel

from genworlds.agents.yeager_autogpt.prompt_generator import get_prompt
from langchain.prompts.chat import (
    BaseChatPromptTemplate,
)
from langchain.schema import BaseMessage, HumanMessage, SystemMessage
from langchain.tools.base import BaseTool
from langchain.vectorstores.base import VectorStoreRetriever
from langchain.vectorstores import Qdrant


class AutoGPTPrompt(BaseChatPromptTemplate, BaseModel):
    ai_name: str
    ai_role: str
    tools: List[BaseTool]
    token_counter: Callable[[str], int]
    send_token_limit: int = 8192

    def construct_full_prompt(self, agent_world_state: str, goals: List[str]) -> str:
        prompt_start = (
            "Your decisions must always be made independently "
            "without seeking user assistance.\n"
            "Play to your strengths as an LLM and pursue simple "
            "strategies with no legal complications.\n"
            "If you have completed all your tasks, make sure to "
            'use the "finish" command.'
            "\n\n"
            "You are an agent that lives in a world with other agents and objects.\n"
            "You can move around the world, interact with objects, and talk to other agents.\n"
            "You have an inventory that can hold objects.\n"
        )

        # Add agent world state
        prompt_start += agent_world_state

        # Construct full prompt
        full_prompt = (
            f"You are {self.ai_name}, {self.ai_role}\n{prompt_start}\n\nGOALS:\n\n"
        )
        for i, goal in enumerate(goals):
            full_prompt += f"{i+1}. {goal}\n"

        full_prompt += f"\n\n{get_prompt(self.tools)}"
        return full_prompt

    def format_messages(self, **kwargs: Any) -> List[BaseMessage]:
        messages: List[BaseMessage] = []

        base_prompt = SystemMessage(
            content=self.construct_full_prompt(
                kwargs["agent_world_state"], kwargs["goals"]
            )
        )
        messages.append(base_prompt)

        time_prompt = SystemMessage(
            content=f"The current time and date is {time.strftime('%c')}"
        )
        messages.append(time_prompt)
        used_tokens = self.token_counter(base_prompt.content) + self.token_counter(
            time_prompt.content
        )

        inventory = kwargs["inventory"]
        if len(inventory) > 0:
            inventory_prompt = f"You have the following items in your inventory:\n"
            for entity in inventory:
                inventory_prompt += f"{json.dumps(entity)}\n"
        else:
            inventory_prompt = f"You have no items in your inventory.\n"
        inventory_message = SystemMessage(content=inventory_prompt)
        messages.append(inventory_message)
        used_tokens += self.token_counter(inventory_message.content)

        nearby_entities = kwargs["nearby_entities"]
        if len(nearby_entities) > 0:
            nearby_entities_prompt = f"There are the following entities near you: \n"
            for entity in nearby_entities:
                nearby_entities_prompt += f"{json.dumps(entity)}\n"
        else:
            nearby_entities_prompt = f"There are no entities near you.\n"
        nearby_entities_message = SystemMessage(content=nearby_entities_prompt)
        messages.append(nearby_entities_message)
        used_tokens += self.token_counter(nearby_entities_message.content)

        relevant_commands = kwargs["relevant_commands"]
        relevant_commands_prompt = f'You can perform the following additional commands with the entities nearby. "target_id" is the id of the entity that provides the command:\n'
        for command in relevant_commands:
            relevant_commands_prompt += f"{command}\n"
        relevant_commands_message = SystemMessage(content=relevant_commands_prompt)
        messages.append(relevant_commands_message)
        used_tokens += self.token_counter(relevant_commands_message.content)

        personality_db: Qdrant = kwargs["personality_db"]
        if personality_db is not None:
            previous_messages = kwargs["messages"]
            past_statements = list(
                map(
                    lambda d: d.page_content,
                    personality_db.similarity_search(str(previous_messages[-10:])),
                )
            )
            if len(past_statements) > 0:
                past_statements_format = f"You have said the following things in the past on this topic:\n{past_statements}\n\n"
                personality_message = SystemMessage(content=past_statements_format)
                messages.append(personality_message)
                used_tokens += len(personality_message.content)

        memory: VectorStoreRetriever = kwargs["memory"]
        previous_messages = kwargs["messages"]
        relevant_docs = memory.get_relevant_documents(str(previous_messages[-10:]))
        relevant_memory = [d.page_content for d in relevant_docs]
        relevant_memory_tokens = sum(
            [self.token_counter(doc) for doc in relevant_memory]
        )
        while used_tokens + relevant_memory_tokens > 2500:
            relevant_memory = relevant_memory[:-1]
            relevant_memory_tokens = sum(
                [self.token_counter(doc) for doc in relevant_memory]
            )
        content_format = (
            f"This reminds you of these events "
            f"from your past:\n{relevant_memory}\n\n"
        )
        memory_message = SystemMessage(content=content_format)
        messages.append(memory_message)
        used_tokens += len(memory_message.content)

        # historical_messages: List[BaseMessage] = []
        # for message in previous_messages[-10:][::-1]:
        #     message_tokens = self.token_counter(message.content)
        #     if used_tokens + message_tokens > self.send_token_limit - 1000:
        #         break
        #     historical_messages = [message] + historical_messages
        # input_message = HumanMessage(content=kwargs["user_input"])

        # messages += historical_messages
        # plan: Optional[str] = kwargs["plan"]

        # messages.append(input_message)
        return messages



import json
from textwrap import dedent
from typing import Callable, Optional, Type, TypedDict
from langchain import BasePromptTemplate, PromptTemplate, LLMChain
from langchain.chat_models import ChatOpenAI
from genworlds.agents.tree_agent.brains.brain import Brain
from genworlds.agents.tree_agent.prompts.execution_generator_prompt import (
    ExecutionGeneratorPrompt,
)

from langchain.output_parsers.openai_functions import JsonKeyOutputFunctionsParser, JsonOutputFunctionsParser
from langchain.chains.openai_functions.utils import get_llm_kwargs

class MultiEvalBrain(Brain):
    """This brain generates a number of thoughts, and passes them to the evaluator one by one to be rated.
    It then selects the highest rated thought and returns it, or None if no thought was rated above the threshold.
    """

    def __init__(
        self,
        openai_api_key: str,
        prompt_template_class: Type[BasePromptTemplate],
        llm_params: list[str],
        generator_role_prompt: str,
        generator_results_prompt: str,
        evaluator_role_prompt: str,
        evaluator_results_prompt: str,
        n_of_thoughts: int,
        output_parameter_generator: Callable[[dict], dict[str, dict]],
        value_threshold: float = 0,
        model_name="gpt-4",
        temperature=0.7,
        verbose=False,
        request_timeout=120,
    ):
        self.n_of_thoughts = n_of_thoughts
        self.value_threshold = value_threshold
        self.verbose = verbose

        self.prompt_template_class = prompt_template_class
        self.llm_params = llm_params

        self.output_parameter_generator = output_parameter_generator

        self.llm = ChatOpenAI(
            temperature=temperature,
            openai_api_key=openai_api_key,
            model_name=model_name,
            request_timeout=request_timeout,
        )

        self.gen_prompt = prompt_template_class(
            token_counter=self.llm.get_num_tokens,
            input_variables=[
                "previous_thoughts",  # iterative
                "num_thoughts",  # num
            ]
            + llm_params,
            ai_role=dedent(generator_role_prompt),
            response_instruction=dedent(generator_results_prompt),
        )

        self.eval_prompt = prompt_template_class(
            token_counter=self.llm.get_num_tokens,
            input_variables=[
                "thought_to_evaluate",
            ]
            + llm_params,
            ai_role=dedent(evaluator_role_prompt),
            response_instruction=dedent(evaluator_results_prompt),
        )

    def gen_thoughts(
        self,
        previous_thoughts,
        num_thoughts: int,
        llm_params: dict,
    ):
        generator_function = {
            "name": "generate_options",
            "description": f"Generates {num_thoughts} options for the agent to choose from.",
            "parameters": {
                "type": "object",
                "properties": {
                    "options": {
                        "type": "array",
                        "minItems": num_thoughts,
                        "maxItems": num_thoughts,
                        "items": {
                            "type": "object",
                            "properties": self.output_parameter_generator(llm_params),
                        }
                    },
                },
                "required": ["options"],
            },
        }


        llm_kwargs = get_llm_kwargs(generator_function)
        output_parser = JsonKeyOutputFunctionsParser(key_name="options")

        if self.verbose:
            print(llm_kwargs)

        gen_llm_chain = LLMChain(
            prompt=self.gen_prompt,
            llm=self.llm,
            llm_kwargs=llm_kwargs,
            output_parser=output_parser,
            verbose=self.verbose,
        )

        # prepare the input variables
        response = gen_llm_chain.run(
            num_thoughts=num_thoughts,
            previous_thoughts=previous_thoughts,
            **llm_params,
        )

        if self.verbose:
            print("Generated: " + str(response))

        return response
    
    def eval_thoughts(
        self,
        thoughts_to_evaluate: list,
        llm_params: dict,
    ):
        thought_values = {}

        for thought in thoughts_to_evaluate:
            evaluator_function = {
                "name": "evaluate_output",
                "description": "Evaluate the output of the generator.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "value": {
                            "type": "number",
                        },
                    },                        
                    "required": ["value"],
                },
            }

            llm_kwargs = get_llm_kwargs(evaluator_function)
            output_parser = JsonOutputFunctionsParser()

            eval_llm_chain = LLMChain(
                prompt=self.eval_prompt,
                llm=self.llm,
                llm_kwargs=llm_kwargs,
                output_parser=output_parser,
                verbose=self.verbose,
            )

            response = eval_llm_chain.run(
                thought_to_evaluate=json.dumps(thought),
                **llm_params,
            )

            thought_values[thought] = response["value"]

        if self.verbose:
                print("Evaluated: " + str(thought_values))

        return thought_values

    def run(self, llm_params: dict):
        thoughts = self.gen_thoughts("", self.n_of_thoughts, llm_params)
        if self.n_of_thoughts == 1 and self.value_threshold == 0:
            return thoughts[0]

        thought_values = self.eval_thoughts(thoughts, llm_params)

        if self.verbose:
            print("Evaluated: ", thought_values)

        best_thought = max(thought_values, key=thought_values.get)

        if thought_values[best_thought] < self.value_threshold:
            return None
        else:
            return best_thought



from prompt_toolkit.layout.processors import (
    Processor,
    Transformation,
)
from prompt_toolkit.formatted_text import to_formatted_text
from prompt_toolkit.formatted_text.utils import fragment_list_to_text
from prompt_toolkit.formatted_text import HTML


class FormatText(Processor):
    def apply_transformation(self, ti):
        fragments = to_formatted_text(HTML(fragment_list_to_text(ti.fragments)))
        return Transformation(fragments)


def html_from_json_format(format, content, max_width=100):
    new_lines = []
    while len(content) > max_width:
        new_lines.append(content[0:max_width])
        content = content[max_width:]
    new_lines.append(content)    
    
    new_lines_styled = [apply_style(format, line) for line in new_lines]

    return "\n".join(new_lines_styled)

def apply_style(style, content):
    for el in style.split(" "):
        if el.startswith("fg:"):
            content = f'<style fg="{el[3:]}">{content}</style>'
        elif el.startswith("bg:"):
            content = f'<style bg="{el[3:]}">{content}</style>'
        elif el == "bold":
            content = f"<b>{content}</b>"
        elif el == "italic":
            content = f"<i>{content}</i>"
        elif el == "underline":
            content = f"<u>{content}</u>"
        elif el == "blink":
            content = f"<blink>{content}</blink>"
        elif el == "reverse":
            content = f"<reverse>{content}</reverse>"
        elif el == "strike":
            content = f"<strike>{content}</strike>"
    return content



from genworlds.sockets.world_socket_client import WorldSocketClient
from langchain.schema import Document


class ListeningAntenna:
    special_events: set[str]
    agent_world_state = "You have not yet learned about the world state."
    schemas: dict
    nearby_entities: list

    important_event_types: set[str]

    all_events: list
    last_events: list
    agent_name: str
    agent_id: str

    def __init__(
        self,
        important_event_types: set[str],
        agent_name,
        agent_id,
        websocket_url: str = "ws://127.0.0.1:7456/ws",
    ):
        self.world_socket_client = WorldSocketClient(
            process_event=self.process_event, url=websocket_url
        )

        self.websocket_url = websocket_url

        self.special_events = {
            "world_sends_schemas_event",
            "entity_world_state_update_event",
            "world_sends_nearby_entities_event",
            "world_sends_all_entities_event",
        }
        self.schemas = {}
        self.nearby_entities = []
        self.all_events = []
        self.last_events = []

        self.important_event_types = self.special_events.copy()
        self.important_event_types.update(important_event_types)

        self.agent_name = agent_name
        self.agent_id = agent_id

    def process_event(self, event):
        match event:
            case {"event_type": "world_sends_schemas_event"}:
                self.schemas = event["schemas"]
            case {"event_type": "entity_world_state_update_event", "target_id": self.agent_id}:
                self.agent_world_state = event["entity_world_state"]
            case {"event_type": "world_sends_nearby_entities_event", "target_id": self.agent_id}:
                self.nearby_entities = event["nearby_entities"]
            case {"event_type": event_type} if event_type in self.special_events:
                pass
            case event:
                if (
                    event["sender_id"] != self.agent_id 
                    and
                    (event["target_id"] == self.agent_id or 
                    event["target_id"] == None or
                    event["event_type"] in self.important_event_types)
                ):
                    self.last_events.append(event)
                    self.all_events.append(event)

    def get_last_events(self):
        events_to_return = self.last_events.copy()
        self.last_events = []
        return events_to_return

    def get_all_events(self):
        return self.all_events

    def get_agent_world_state(self):
        return self.agent_world_state

    def get_nearby_entities(self):
        return self.nearby_entities

    def get_schemas(self):
        return self.schemas



from __future__ import annotations

from prompt_toolkit.layout import HSplit
from prompt_toolkit.layout.containers import Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.layout import WindowAlign
from prompt_toolkit.layout.dimension import D
from prompt_toolkit.keys import Keys

import genworlds.interfaces as interfaces
import genworlds.interfaces.cli.render_main_screen as render_main_screen


def render_help_screen(cli: interfaces.CLI):
    title = Window(
        height=D.exact(3),
        content=FormattedTextControl(
            text=HTML('<style fg="#ffb86c">\n🧬🌍 GenWorlds CLI Help Screen\n</style>')
        ),
        align=WindowAlign.CENTER,
    )

    main_container = HSplit(
        [
            Window(
                height=1,
                content=FormattedTextControl(
                    text="Press ESC to return to the main screen"
                ),
                align=WindowAlign.CENTER,
            ),
            Window(
                height=1,
                content=FormattedTextControl(
                    text="Press CTRL+W to write buffers to disk"
                ),
                align=WindowAlign.CENTER,
            ),
            Window(
                height=1,
                content=FormattedTextControl(
                    text="Press CTRL+H to view this help screen"
                ),
                align=WindowAlign.CENTER,
            ),
            Window(
                height=1,
                content=FormattedTextControl(text="Press CTRL+C to exit"),
                align=WindowAlign.CENTER,
            ),
        ],
    )

    # Update the CLI layout
    cli.application.layout.container = HSplit(
        [
            title,
            main_container,
        ],
    )

    # Update the CLI keybindings
    @cli.kb.add(Keys.Escape, eager=True)
    def _(event):
        render_main_screen.render_main_screen(cli)



from uuid import uuid4
from genworlds.events.websocket_event_handler import WebsocketEventHandler
from genworlds.events.basic_events import (
    AgentGetsObjectInfoEvent,
    ObjectSendsInfoToAgentEvent,
)


class BaseObject(WebsocketEventHandler):
    def __init__(
        self,
        name: str,
        description: str,
        id: str = None,
        websocket_url: str = "ws://127.0.0.1:7456/ws",
    ):
        self.id = id if id else str(uuid4())
        self.name = name
        self.description = description

        super().__init__(
            id=self.id,
            websocket_url=websocket_url,
        )
        self.register_event_listeners(
            [
                # (AgentGetsObjectInfoEvent, self.agent_gets_object_info_listener)
            ]
        )

    def agent_gets_object_info_listener(self, event: AgentGetsObjectInfoEvent):
        self.send_event(
            ObjectSendsInfoToAgentEvent,
            target_id=event.agent_id,
            object_name=self.name,
            object_description=self.description,
            possible_events=self.event_classes.keys(),
        )



from __future__ import annotations

from datetime import datetime
import threading
from uuid import uuid4
from time import sleep
import json
from typing import List, Optional

from pydantic import ValidationError
from jsonschema import validate

from langchain.tools import StructuredTool
from langchain.tools.human.tool import HumanInputRun
from langchain.vectorstores.base import VectorStoreRetriever
from langchain.vectorstores import Qdrant
from langchain.embeddings import OpenAIEmbeddings
from langchain.schema import (
    AIMessage,
    BaseMessage,
    Document,
    HumanMessage,
    SystemMessage,
)
from qdrant_client import QdrantClient
from genworlds.agents.tree_agent.brains.brain import Brain

from genworlds.events.basic_events import (
    EntityRequestWorldStateUpdateEvent,
)
from genworlds.utils.logging_factory import LoggingFactory

from genworlds.sockets.world_socket_client import WorldSocketClient
from genworlds.agents.world_listeners.listening_antenna import ListeningAntenna
from genworlds.agents.memory_processors.nmk_world_memory import NMKWorldMemory

FINISH_NAME = "finish"


class TreeAgent:
    """Agent class for structured tree-of-thought execution."""

    world_spawned_id: str
    personality_db = None

    def __init__(
        self,
        name: str,
        description: str,
        goals: List[str],
        openai_api_key: str,
        navigation_brain: Brain,
        execution_brains: dict[str, Brain],
        action_brain_map: dict,
        interesting_events: set = {},
        can_sleep: bool = True,
        wakeup_events: dict = {},
        feedback_tool: Optional[HumanInputRun] = None,
        additional_memories: Optional[List[VectorStoreRetriever]] = None,
        id: str = None,
        personality_db_qdrant_client: QdrantClient = None,
        personality_db_collection_name: str = None,
        websocket_url: str = "ws://127.0.0.1:7456/ws",
    ):
        # Its own properties
        self.id = id if id else str(uuid4())
        self.name = name
        self.description = description
        self.goals = goals
        self.interesting_events = interesting_events
        self.can_sleep = can_sleep
        self.wakeup_events = wakeup_events
        self.feedback_tool = feedback_tool
        self.additional_memories = additional_memories

        self.is_asleep = False

        self.logger = LoggingFactory.get_logger(self.name)

        self.world_socket_client = WorldSocketClient(
            process_event=None, url=websocket_url
        )

        self.listening_antenna = ListeningAntenna(
            self.interesting_events,
            agent_name=self.name,
            agent_id=self.id,
            websocket_url=websocket_url,
        )

        # Agent actions
        self.actions = []

        # Brain properties
        self.nmk_world_memory = NMKWorldMemory(
            openai_api_key=openai_api_key,
            n_of_last_events=10,
            n_of_similar_events=0,
            n_of_paragraphs_in_summary=3,
        )

        self.navigation_brain = navigation_brain
        self.execution_brains = execution_brains
        self.action_brain_map = action_brain_map

        self.full_message_history: List[BaseMessage] = []
        self.next_action_count = 0
        self.feedback_tool = None  # HumanInputRun() if human_in_the_loop else None
        self.schemas_memory: Qdrant
        self.plan: Optional[str] = None

        self.embeddings_model = OpenAIEmbeddings(openai_api_key=openai_api_key)
        self.personality_db_qdrant_client = personality_db_qdrant_client
        if self.personality_db_qdrant_client:
            self.personality_db = Qdrant(
                collection_name=personality_db_collection_name,
                embeddings=self.embeddings_model,
                client=self.personality_db_qdrant_client,
            )
            
    def think(self):
        self.logger.info(f" The agent {self.name} is thinking...")
        user_input = (
            "Determine which next command to use, "
            "and respond using the format specified above:"
        )
        # Get the initial world state
        self.agent_request_world_state_update_action()

        sleep(5)

        while True:
            # If there are any relevant events in the world for this agent, add them to memory
            last_events = self.listening_antenna.get_last_events()
            self.logger.debug(f"Last events: {last_events}")
            for event in last_events:
                self.nmk_world_memory.add_event(json.dumps(event), summarize=True)
                self.handle_wakeup(event)

            if self.is_asleep:
                self.logger.info(f"Sleeping...")
                sleep(5)
                continue

            agent_world_state = self.listening_antenna.get_agent_world_state()
            nearby_entities = self.listening_antenna.get_nearby_entities()

            if len(nearby_entities) > 0:
                nearby_entities_store = Qdrant.from_texts(
                    list(map(json.dumps, nearby_entities)),
                    self.embeddings_model,
                    location=":memory:",
                )

                if self.plan:
                    useful_nearby_entities = nearby_entities_store.similarity_search(
                        str(self.plan)
                    )
                else:
                    useful_nearby_entities = nearby_entities_store.similarity_search(
                        json.dumps(self.goals)
                    )

                useful_nearby_entities = list(
                    map(lambda d: json.loads(d.page_content), useful_nearby_entities)
                )
            else:
                useful_nearby_entities = []

            relevant_commands = {
                # Default commands
                "Self:wait": {
                    "title": "Self:wait",
                    "description": "Wait until the state of the world changes",
                    "args": {},
                    "string_short": "Self:wait - Wait until the state of the world changes",
                    "string_full": "Self:wait - Wait until the state of the world changes, args json schema: {}",
                },
            }
            for entity in useful_nearby_entities:
                entity_schemas = self.get_schemas()[entity["entity_class"]]

                for event_type, schema in entity_schemas.items():
                    description = schema["properties"]["description"]["default"]

                    args = {}
                    for property_name, property_details in schema["properties"].items():
                        if property_name not in [
                            "event_type",
                            "description",
                            "created_at",
                            "sender_id",
                            "summary",
                        ]:
                            args[property_name] = property_details

                    selected_command = {
                        "title": f"{entity['entity_class']}:{event_type}",
                        "description": description,
                        "args": args,
                        "string_short": f"{entity['entity_class']}:{event_type} - {description}",
                        "string_full": f"\"{entity['entity_class']}:{event_type}\" - {description}, args json schema: {json.dumps(args)}",
                    }
                    relevant_commands[selected_command["title"]] = selected_command

            # Add world
            entity_class = "World"
            entity_schemas = self.get_schemas()[entity_class]

            for event_type, schema in entity_schemas.items():
                if event_type in self.listening_antenna.special_events:
                    continue

                description = schema["properties"]["description"]["default"]

                args = {}
                for property_name, property_details in schema["properties"].items():
                    if property_name not in [
                        "event_type",
                        "description",
                        "created_at",
                        "sender_id",
                        "summary",
                    ]:
                        args[property_name] = property_details

                selected_command = {
                    "title": f"{entity_class}:{event_type}",
                    "description": description,
                    "args": args,
                    "string_short": f"{entity_class}:{event_type} - {description}",
                    "string_full": f'"{entity_class}:{event_type}" - {description}, args json schema: {json.dumps(args)}',
                }
                relevant_commands[selected_command["title"]] = selected_command

            # Send message to AI, get response
            navigation_plan_parsed = self.navigation_brain.run(
                {
                    "goals": self.goals,
                    "messages": self.full_message_history,
                    "memory": self.nmk_world_memory,
                    "personality_db": self.personality_db,
                    "nearby_entities": list(
                        filter(lambda e: (e["held_by"] != self.id), nearby_entities)
                    ),
                    "inventory": list(
                        filter(lambda e: (e["held_by"] == self.id), nearby_entities)
                    ),
                    "plan": self.plan,
                    "user_input": user_input,
                    "agent_world_state": agent_world_state,
                    "relevant_commands": relevant_commands,
                }
            )

            # Print Assistant thoughts
            self.logger.info(navigation_plan_parsed)
            self.full_message_history.append(
                AIMessage(content=str(navigation_plan_parsed))
            )

            self.plan = navigation_plan_parsed["plan"]

            selected_action = navigation_plan_parsed["next_action"]
            action_goal_description = navigation_plan_parsed["next_action_aim"]

            result = ""
            event_sent_summary = ""
            if selected_action == "Self:exit":
                self.logger.info(f"Exiting...")
                return "FINISHED"
            elif selected_action == "Self:wait":
                self.logger.info(f"Waiting, entering sleep mode...")
                if self.can_sleep:
                    self.is_asleep = True
                else:
                    sleep(10)
                result += f"Waiting...\n"
            # TODO: tools?
            elif selected_action in relevant_commands:
                selected_command = relevant_commands[selected_action]

                if selected_action in self.action_brain_map:
                    action_brains = self.action_brain_map[selected_action]
                else:
                    action_brains = self.action_brain_map["default"]

                previous_brain_outputs = [
                    f"Current goal: {action_goal_description}",
                ]
                for action_brain_name in action_brains:
                    action_brain = self.execution_brains[action_brain_name]

                    previous_brain_outputs.append(
                        action_brain.run(
                            {
                                "goals": self.goals,
                                "messages": self.full_message_history,
                                "memory": self.nmk_world_memory,
                                "personality_db": self.personality_db,
                                "nearby_entities": list(
                                    filter(
                                        lambda e: (e["held_by"] != self.id),
                                        nearby_entities,
                                    )
                                ),
                                "inventory": list(
                                    filter(
                                        lambda e: (e["held_by"] == self.id),
                                        nearby_entities,
                                    )
                                ),
                                "plan": self.plan,
                                "user_input": user_input,
                                "agent_world_state": agent_world_state,
                                "command_to_execute": selected_command,
                                "previous_brain_outputs": previous_brain_outputs,
                            }
                        )
                    )
                final_brain_output = previous_brain_outputs[-1]
                try:
                    args = final_brain_output

                    if type(args) == dict:
                        event_sent = self.execute_event_with_args(
                            selected_command["title"], args
                        )
                        self.nmk_world_memory.add_event(
                            json.dumps(event_sent), summarize=True
                        )
                        event_sent_summary += (
                            "Event at: " + event_sent["created_at"] + "\n"
                        )
                        event_sent_summary += (
                            "What happened: " + str(event_sent["summary"]) + "\n"
                        )

                        result += event_sent_summary
                    else:
                        raise Exception("Unexpected final output")

                except Exception as e:
                    self.logger.error(
                        f"Problem executing command with {selected_command['title']} with output {final_brain_output}: {e}\n"
                    )
                    result += f"Problem executing command with {selected_command['title']} with output {final_brain_output}: {e}\n"
            else:
                self.logger.info(f"Invalid command: {selected_action}")
                result += f"Error: {selected_action} is not recognized. \n"
                continue

            ## send result and assistant_reply to the socket
            self.logger.info(result)

            self.full_message_history.append(SystemMessage(content=result))

            # Allow events to be processed
            sleep(3)

    def get_agent_world_state(self):
        return self.listening_antenna.get_agent_world_state()

    def get_nearby_entities(self):
        return self.listening_antenna.get_nearby_entities()

    def get_schemas(self):
        return self.listening_antenna.get_schemas()

    def execute_event_with_args(self, name: str, args: dict):
        try:
            class_name = name.split(":")[0]
            event_type = name.split(":")[1]

            event = {
                "event_type": event_type,
                "sender_id": self.id,
                "created_at": datetime.now().isoformat(),
            }
            event.update(args)
            summary = self.nmk_world_memory.one_line_summarizer.summarize(
                json.dumps(event)
            )
            event["summary"] = summary
            self.logger.debug(event)

            event_schema = self.get_schemas()[class_name][event_type]
            validate(event, event_schema)

            self.world_socket_client.send_message(json.dumps(event))
            return event
        except IndexError as e:
            return (
                f"Unknown command '{name}'. "
                f"Please refer to the 'COMMANDS' list for available "
                f"commands and only respond in the specified JSON format."
            )
        except ValidationError as e:
            return f"Validation Error in args: {str(e)}, args: {args}"
        except Exception as e:
            return f"Error: {str(e)}, {type(e).__name__}, args: {args}"

    def agent_request_world_state_update_action(self):
        agent_request_world_state_update = EntityRequestWorldStateUpdateEvent(
            created_at=datetime.now(),
            sender_id=self.id,
            target_id=self.world_spawned_id,
        )
        self.world_socket_client.send_message(agent_request_world_state_update.json())

    def handle_wakeup(self, event):
        event_type = event["event_type"]
        if event_type not in self.wakeup_events:
            return

        wakeup_event_property_filters = self.wakeup_events[event_type]

        # Check if the event matches the filters
        is_match = True
        for wakeup_event_property in wakeup_event_property_filters.keys():
            expected_value = wakeup_event_property_filters[wakeup_event_property]
            if event[wakeup_event_property] != expected_value:
                is_match = False
                break

        if is_match:
            self.logger.info("Waking up", event)
            self.is_asleep = False

    def launch_threads(self):
        threading.Thread(
            target=self.listening_antenna.world_socket_client.websocket.run_forever,
            name=f"Agent {self.name} Listening Thread",
            daemon=True,
        ).start()
        sleep(0.1)
        threading.Thread(
            target=self.world_socket_client.websocket.run_forever,
            name=f"Agent {self.name} Speaking Thread",
            daemon=True,
        ).start()
        sleep(0.1)
        threading.Thread(
            target=self.think,
            name=f"Agent {self.name} Thinking Thread",
            daemon=True,
        ).start()
        self.logger.info("Threads launched")



from genworlds.events.websocket_event_handler import Event
from genworlds.worlds.base_world import BaseWorld
from genworlds.worlds.base_world_entity import BaseWorldEntity


class WorldEntity2D(BaseWorldEntity):
    location: str = None


class AgentMovesToNewLocation(Event):
    event_type = "agent_moves_to_new_location"
    description = "Agent moves to a new location in the world."
    destination_location: str


class World2D(BaseWorld[WorldEntity2D]):
    locations: list[str] = []

    def __init__(
        self,
        name: str,
        description: str,
        locations: list[str],
        id: str = None,
        websocket_url: str = "ws://127.0.0.1:7456/ws",
    ):
        super().__init__(
            world_entity_constructor=WorldEntity2D,
            name=name,
            description=description,
            id=id,
            websocket_url=websocket_url,
        )

        self.locations = locations

        self.register_event_listeners(
            [
                (AgentMovesToNewLocation, self.agent_moves_to_new_location_listener),
            ]
        )

    def get_nearby_entities(self, entity_id: str) -> WorldEntity2D:
        reference_entity = self.entities.get(entity_id)

        nearby_entities = []
        # Same location or in my inventory
        for entity in self.entities.values():
            if entity != reference_entity:
                if (
                    entity.location == reference_entity.location
                    or entity.held_by == reference_entity.id
                ):
                    nearby_entities.append(entity)

        # Add objects in inventory of nearby entities
        for entity in self.entities.values():
            if entity.held_by in [e.id for e in nearby_entities]:
                nearby_entities.append(entity)

        # All entities
        return nearby_entities

    def get_agent_world_state_prompt(self, agent_id: str) -> str:
        agent_entity = self.get_agent_by_id(agent_id)

        world_state_prompt = (
            f"You are an agent in a 2D world.\n"
            f'THe id of the world is "{self.id}".\n'
            f'Your id is "{agent_entity.id}".\n'
            f'Your location is "{agent_entity.location}".\n'
            f'Available locations are "{self.locations}".\n'
        )

        return world_state_prompt

    def agent_moves_to_new_location_listener(self, event: AgentMovesToNewLocation):
        agent_entity = self.get_agent_by_id(event.agent_id)

        if event.destination_location not in self.locations:
            raise Exception(f"Location {event.destination_location} does not exist.")

        agent_entity.location = event.destination_location
        self.emit_agent_world_state(agent_entity.id)



import time
from typing import Literal


class TreeOfThoughts:
    def __init__(
        self,
        gen_thoughts: callable,
        eval_thoughts: callable,
        search_algorithm: Literal["BFS", "DFS"],
        initial_state: str,
        thought_limit: int,
        max_depth: int,
        breadth: int,
        value_threshold: int,
        timeout: int = None,
    ):
        self.gen_thoughts = gen_thoughts
        self.eval_thoughts = eval_thoughts
        self.search_algorithm = search_algorithm
        self.initial_state = initial_state
        self.thought_limit = thought_limit
        self.max_depth = max_depth
        self.breadth = breadth
        self.value_threshold = value_threshold
        self.timeout = timeout

    def solve(
        self,
        llm_params: dict,
    ) -> list[str]:
        start_time = time.time()

        if self.search_algorithm == "BFS":
            return self.bfs_until_timeout(
                self.initial_state,
                self.thought_limit,
                self.max_depth,
                self.breadth,
                start_time,
                self.timeout,
                llm_params,
            )
        elif self.search_algorithm == "DFS":
            return self.dfs_until_timeout(
                self.initial_state,
                self.thought_limit,
                self.max_depth,
                self.value_threshold,
                start_time,
                self.timeout,
                llm_params,
            )
        else:
            raise ValueError("Invalid search algorithm. Choose 'BFS' or 'DFS'.")

    def bfs_until_timeout(
        self,
        initial_state,
        thought_limit,
        max_depth,
        breadth,
        start_time,
        timeout,
        llm_params,
    ):
        while timeout is None or time.time() - start_time < timeout:
            result = self.tot_bfs(
                initial_state, thought_limit, max_depth, breadth, llm_params
            )
            if result:
                return result

    def tot_bfs(self, initial_state, thought_limit, max_depth, breadth, llm_params):
        current_set = {initial_state}
        for _ in range(max_depth):
            current_set = self.expand_set(
                current_set, thought_limit, breadth, llm_params
            )
        print(current_set)
        print(max(current_set))
        # return self.gen_thoughts(max(current_set), 1, llm_params)
        return max(current_set)

    def expand_set(self, current_set, thought_limit, breadth, llm_params):
        new_set = {
            (*state, new_thought)
            for state in current_set
            for new_thought in self.gen_thoughts(state, thought_limit, llm_params)
        }
        if thought_limit == 1:
            return new_set
        else:
            thought_values = self.eval_thoughts(new_set, llm_params)
            sorted_set = sorted(
                new_set, key=lambda state: thought_values[state], reverse=True
            )
            return set(sorted_set[:breadth])

    def dfs_until_timeout(
        self,
        initial_state,
        thought_limit,
        max_depth,
        value_threshold,
        start_time,
        timeout,
        llm_params,
    ):
        while timeout is None or time.time() - start_time < timeout:
            result = self.tot_dfs(
                initial_state, thought_limit, max_depth, value_threshold, llm_params
            )
            if result:
                return result

    def tot_dfs(
        self,
        initial_state,
        thought_limit,
        max_depth,
        value_threshold,
        llm_params,
        pruning_threshold=0.5,
        confidence_threshold=0.9,
        max_iterations=10,
        convergence_threshold=0.1,
        convergence_count=5,
    ):
        output, iteration_count, consecutive_convergence_count, prev_best_value = (
            [],
            0,
            0,
            None,
        )
        self.dfs_search(
            initial_state,
            1,
            max_depth,
            thought_limit,
            value_threshold,
            pruning_threshold,
            confidence_threshold,
            output,
            iteration_count,
            consecutive_convergence_count,
            prev_best_value,
            max_iterations,
            convergence_threshold,
            convergence_count,
            llm_params,
        )
        return max(output, key=lambda output_item: output_item[1]) if output else None

    def dfs_search(
        self,
        current_state,
        current_depth,
        max_depth,
        thought_limit,
        value_threshold,
        pruning_threshold,
        confidence_threshold,
        output,
        iteration_count,
        consecutive_convergence_count,
        prev_best_value,
        max_iterations,
        convergence_threshold,
        convergence_count,
        llm_params,
    ):
        if current_depth > max_depth or self.should_terminate(
            output,
            iteration_count,
            consecutive_convergence_count,
            max_iterations,
            convergence_count,
            confidence_threshold,
            convergence_threshold,
            prev_best_value,
        ):
            return
        for next_state in self.gen_thoughts(current_state, thought_limit, llm_params):
            state_value = self.eval_thoughts({next_state}, llm_params)[next_state]
            if state_value > value_threshold and (
                pruning_threshold is None or state_value >= pruning_threshold
            ):
                self.dfs_search(
                    (*current_state, next_state),
                    current_depth + 1,
                    max_depth,
                    thought_limit,
                    value_threshold,
                    pruning_threshold,
                    confidence_threshold,
                    output,
                    iteration_count,
                    consecutive_convergence_count,
                    prev_best_value,
                    max_iterations,
                    convergence_threshold,
                    convergence_count,
                    llm_params,
                )

    def should_terminate(
        self,
        output,
        iteration_count,
        consecutive_convergence_count,
        max_iterations,
        convergence_count,
        confidence_threshold,
        convergence_threshold,
        prev_best_value,
    ):
        if (
            iteration_count >= max_iterations
            or consecutive_convergence_count >= convergence_count
        ):
            return True
        if output:
            thought, value = output[-1]
            if value >= confidence_threshold or (
                prev_best_value is not None
                and abs(value - prev_best_value) < convergence_threshold
            ):
                return True
        return False



from __future__ import annotations
from datetime import datetime
import threading
from uuid import uuid4
from time import sleep
import json
from typing import List, Optional

from pydantic import ValidationError

from jsonschema import validate

from langchain.chat_models import ChatOpenAI
from langchain.docstore import InMemoryDocstore
from langchain.tools import StructuredTool
from langchain.tools.human.tool import HumanInputRun
from langchain.vectorstores.base import VectorStoreRetriever

from langchain.vectorstores import Qdrant
from qdrant_client import QdrantClient
from genworlds.agents.memory_processors.nmk_world_memory import NMKWorldMemory

from langchain.embeddings import OpenAIEmbeddings
from langchain.chains.llm import LLMChain
from langchain.schema import (
    AIMessage,
    BaseMessage,
    Document,
    HumanMessage,
    SystemMessage,
)
from genworlds.agents.world_listeners.listening_antenna import ListeningAntenna

from genworlds.agents.yeager_autogpt.output_parser import (
    AutoGPTAction,
    AutoGPTOutputParser,
)
from genworlds.agents.yeager_autogpt.prompt import AutoGPTPrompt
from genworlds.agents.yeager_autogpt.prompt_generator import FINISH_NAME
from genworlds.sockets.world_socket_client import WorldSocketClient
from genworlds.agents.yeager_autogpt.memory_summarizers import MemorySummarizer
from genworlds.events.basic_events import (
    AgentGetsNearbyEntitiesEvent,
    AgentGetsObjectInfoEvent,
    AgentGetsAgentInfoEvent,
    AgentGivesObjectToAgentEvent,
    AgentSpeaksWithAgentEvent,
    EntityRequestWorldStateUpdateEvent,
)
from genworlds.utils.logging_factory import LoggingFactory


class YeagerAutoGPT:
    """Agent class for interacting with Auto-GPT."""

    world_spawned_id: str
    personality_db = None

    def __init__(
        self,
        ai_name: str,
        description: str,
        goals: List[str],
        openai_api_key: str,
        interesting_events: set = {},
        feedback_tool: Optional[HumanInputRun] = None,
        additional_memories: Optional[List[VectorStoreRetriever]] = None,
        id: str = None,
        personality_db_qdrant_client: QdrantClient = None,
        personality_db_collection_name: str = None,
        websocket_url: str = "ws://127.0.0.1:7456/ws",
    ):
        # Its own properties
        self.id = id if id else str(uuid4())
        self.ai_name = ai_name
        self.description = description
        self.goals = goals
        self.interesting_events = interesting_events

        self.logger = LoggingFactory.get_logger(self.ai_name)

        self.world_socket_client = WorldSocketClient(
            process_event=None, url=websocket_url
        )

        self.listening_antenna = ListeningAntenna(
            self.interesting_events,
            agent_name=self.ai_name,
            agent_id=self.id,
            websocket_url=websocket_url,
        )

        # Agent actions
        self.actions = []

        # Brain properties
        self.nmk_world_memory = NMKWorldMemory(
            openai_api_key=openai_api_key,
            n_of_last_events=10,
            n_of_similar_events=0,
            n_of_paragraphs_in_summary=3,
        )

        llm = ChatOpenAI(openai_api_key=openai_api_key, model_name="gpt-4")
        prompt = AutoGPTPrompt(
            ai_name=self.ai_name,
            ai_role=self.description,
            tools=self.actions,
            input_variables=[
                "memory",
                "personality_db",
                "messages",
                "goals",
                "user_input",
                "nearby_entities",
                "inventory",
                "relevant_commands",
                "plan",
                "agent_world_state",
            ],
            token_counter=llm.get_num_tokens,
        )
        self.chain = LLMChain(llm=llm, prompt=prompt, verbose=True)

        self.full_message_history: List[BaseMessage] = []
        self.next_action_count = 0
        self.output_parser = AutoGPTOutputParser()
        self.feedback_tool = None  # HumanInputRun() if human_in_the_loop else None
        self.schemas_memory: Qdrant
        self.plan: Optional[str] = None

        self.personality_db_qdrant_client = personality_db_qdrant_client
        if self.personality_db_qdrant_client:
            self.personality_db = Qdrant(
                collection_name=personality_db_collection_name,
                embedding_function=self.embeddings_model,
                client=self.personality_db_qdrant_client,
            )

    def think(self):
        self.logger.info(f" The agent {self.ai_name} is thinking...")
        user_input = (
            "Determine which next command to use, "
            "and respond using the format specified above:"
        )
        # Get the initial world state
        self.agent_request_world_state_update_action()

        sleep(5)

        while True:
            agent_world_state = self.listening_antenna.get_agent_world_state()
            nearby_entities = self.listening_antenna.get_nearby_entities()

            if len(nearby_entities) > 0:
                nearby_entities_store = Qdrant.from_texts(
                    list(map(json.dumps, nearby_entities)), self.embeddings_model
                )

                if self.plan:
                    useful_nearby_entities = nearby_entities_store.similarity_search(
                        self.plan
                    )
                else:
                    useful_nearby_entities = nearby_entities_store.similarity_search(
                        json.dumps(self.goals)
                    )

                useful_nearby_entities = list(
                    map(lambda d: json.loads(d.page_content), useful_nearby_entities)
                )
            else:
                useful_nearby_entities = []

            relevant_commands = []
            for entity in useful_nearby_entities:
                entity_schemas = self.get_schemas()[entity["entity_class"]]

                for event_type, schema in entity_schemas.items():
                    description = schema["properties"]["description"]["default"]

                    args = {}
                    for property_name, property_details in schema["properties"].items():
                        if property_name not in [
                            "event_type",
                            "description",
                            "created_at",
                            "sender_id",
                        ]:
                            args[property_name] = property_details

                    # get_object_info: get_object_info(object_id: 'str') - Gets the info of an object., args json schema: {"object_id": {"title": "Object Id", "type": "string"}}
                    command = f"\"{entity['entity_class']}:{event_type}\" - {description}, args json schema: {json.dumps(args)}"
                    relevant_commands.append(command)

            # Add world
            entity_class = "World"
            entity_schemas = self.get_schemas()[entity_class]

            for event_type, schema in entity_schemas.items():
                if event_type in self.listening_antenna.special_events:
                    continue

                description = schema["properties"]["description"]["default"]

                args = {}
                for property_name, property_details in schema["properties"].items():
                    if property_name not in [
                        "event_type",
                        "description",
                        "created_at",
                        "sender_id",
                    ]:
                        args[property_name] = property_details
                        # args_string += f"{property_name}: {property_details['type']}, "

                # get_object_info: get_object_info(object_id: 'str') - Gets the info of an object., args json schema: {"object_id": {"title": "Object Id", "type": "string"}}
                command = f'"{entity_class}:{event_type}" - {description}, args json schema: {json.dumps(args)}'
                relevant_commands.append(command)

            # Send message to AI, get response
            assistant_reply = self.chain.run(
                goals=self.goals,
                messages=self.full_message_history,
                memory=self.memory,
                personality_db=self.personality_db,
                nearby_entities=list(
                    filter(lambda e: (e["held_by"] != self.id), nearby_entities)
                ),
                inventory=list(
                    filter(lambda e: (e["held_by"] == self.id), nearby_entities)
                ),
                relevant_commands=relevant_commands,
                plan=self.plan,
                user_input=user_input,
                agent_world_state=agent_world_state,
            )
            self.plan = json.loads(assistant_reply)["thoughts"]["plan"]
            # Print Assistant thoughts
            self.logger.info(assistant_reply)  # Send the thoughts as events
            self.full_message_history.append(HumanMessage(content=user_input))
            self.full_message_history.append(AIMessage(content=assistant_reply))

            # Get command name and arguments
            actions = self.output_parser.parse(assistant_reply)
            result = ""
            event_sent_summary = ""
            for action in actions:
                tools = {t.name: t for t in self.actions}
                if action.name == FINISH_NAME:
                    return action.args["response"]
                elif action.name == "ERROR":
                    result += f"Error: {action.args}. \n"
                elif action.name in tools:
                    tool = tools[action.name]
                    try:
                        observation = tool.run(action.args)
                    except ValidationError as e:
                        observation = (
                            f"Validation Error in args: {str(e)}, args: {action.args}"
                        )
                    except Exception as e:
                        observation = (
                            f"Error: {str(e)}, {type(e).__name__}, args: {action.args}"
                        )
                    result += f"Command {tool.name} returned: {observation} \n"
                else:
                    event_sent = self.execute_event_action(action)
                    event_sent_summary += (
                        "Event timestamp: " + event_sent["created_at"] + "\n"
                    )
                    event_sent_summary += event_sent["sender_id"] + " sent "
                    event_sent_summary += event_sent["event_type"] + " to "
                    event_sent_summary += str(event_sent["target_id"]) + "\n"
                    event_sent_summary += (
                        "And this is the summary of what happened: "
                        + str(event_sent["summary"])
                        + "\n"
                    )

            # If there are any relevant events in the world for this agent, add them to memory
            sleep(3)
            last_events = self.listening_antenna.get_last_events()
            memory_to_add = ""
            for event in last_events:
                memory_to_add += "Event timestamp: " + event["created_at"] + "\n"
                memory_to_add += event["sender_id"] + " sent "
                memory_to_add += event["event_type"] + " to "
                memory_to_add += str(event["target_id"]) + "\n"
                memory_to_add += (
                    "And this is the summary of what happened: "
                    + str(event["summary"])
                    + "\n"
                )

            # memory_to_add += event_sent_summary

            self.logger.debug(f"Adding to memory: {memory_to_add}")

            if self.feedback_tool is not None:
                feedback = f"\n{self.feedback_tool.run('Input: ')}"
                if feedback in {"q", "stop"}:
                    self.logger.info("EXITING")
                    return "EXITING"
                memory_to_add += feedback
            if memory_to_add != "":
                self.memory.add_documents([Document(page_content=memory_to_add)])
            self.full_message_history.append(SystemMessage(content=event_sent_summary))

    def get_agent_world_state(self):
        return self.listening_antenna.get_agent_world_state()

    def get_nearby_entities(self):
        return self.listening_antenna.get_nearby_entities()

    def get_schemas(self):
        return self.listening_antenna.get_schemas()

    def execute_event_action(self, action: AutoGPTAction):
        try:
            class_name = action.name.split(":")[0]
            event_type = action.name.split(":")[1]

            event = {
                "event_type": event_type,
                "sender_id": self.id,
                "created_at": datetime.now().isoformat(),
            }
            event.update(action.args)
            summary = self.memory_summarizer.summarize(json.dumps(event))
            event["summary"] = summary
            self.logger.debug(event)

            event_schema = self.get_schemas()[class_name][event_type]
            validate(event, event_schema)

            self.world_socket_client.send_message(json.dumps(event))
            return event
        except IndexError as e:
            return (
                f"Unknown command '{action.name}'. "
                f"Please refer to the 'COMMANDS' list for available "
                f"commands and only respond in the specified JSON format."
            )
        except ValidationError as e:
            return f"Validation Error in args: {str(e)}, args: {action.args}"
        except Exception as e:
            return f"Error: {str(e)}, {type(e).__name__}, args: {action.args}"

    def agent_gets_nearby_entities_action(self):
        agent_gets_nearby_entities_event = AgentGetsNearbyEntitiesEvent(
            created_at=datetime.now(),
            sender_id=self.id,
        )
        self.world_socket_client.send_message(agent_gets_nearby_entities_event.json())

    def agent_gets_object_info_action(
        self,
        target_id: str,
    ):
        agent_gets_object_info = AgentGetsObjectInfoEvent(
            created_at=datetime.now(),
            sender_id=self.id,
            target_id=target_id,
        )
        self.world_socket_client.send_message(agent_gets_object_info.json())

    def agent_gets_agent_info_action(
        self,
        target_id: str,
    ):
        agent_gets_agent_info = AgentGetsAgentInfoEvent(
            created_at=datetime.now(),
            sender_id=self.id,
            target_id=target_id,
        )
        self.world_socket_client.send_message(agent_gets_agent_info.json())

    def agent_interacts_with_object_action(
        self,
        created_interaction: str,
    ):
        self.world_socket_client.send_message(created_interaction)

    def agent_speaks_with_agent_action(
        self,
        target_id: str,
        message: str,
    ):
        agent_speaks_with_agent = AgentSpeaksWithAgentEvent(
            created_at=datetime.now(),
            sender_id=self.id,
            target_id=target_id,
            message=message,
        )
        self.world_socket_client.send_message(agent_speaks_with_agent.json())

    def agent_request_world_state_update_action(self):
        agent_request_world_state_update = EntityRequestWorldStateUpdateEvent(
            created_at=datetime.now(),
            sender_id=self.id,
            target_id=self.world_spawned_id,
        )
        self.world_socket_client.send_message(agent_request_world_state_update.json())

    def launch_threads(self):
        threading.Thread(
            target=self.listening_antenna.world_socket_client.websocket.run_forever,
            name=f"Agent {self.ai_name} Listening Thread",
            daemon=True,
        ).start()
        sleep(0.1)
        threading.Thread(
            target=self.world_socket_client.websocket.run_forever,
            name=f"Agent {self.ai_name} Speaking Thread",
            daemon=True,
        ).start()
        sleep(0.1)
        threading.Thread(
            target=self.think,
            name=f"Agent {self.ai_name} Thinking Thread",
            daemon=True,
        ).start()
        self.logger.info("Threads launched")
